"use strict";

var JSONPath, chai, common, debug, extractMatches, findOperator, operators;
debug = require('debug')('fbp-spec:expectation');
chai = require('chai');
JSONPath = require('JSONPath');
common = require('./common'); //# A library of standard expectation operators
// Should throw AssertionError if expectation is violated,
// with details about the issue

operators = {
  'type': function type(actual, expected) {
    return chai.expect(actual).to.be.a(expected);
  },
  'equals': function equals(actual, expected) {
    return chai.expect(actual).to.eql(expected);
  },
  'above': function above(actual, expected) {
    return chai.expect(actual).to.be.above(expected);
  },
  'below': function below(actual, expected) {
    return chai.expect(actual).to.be.below(expected);
  },
  'haveKeys': function haveKeys(actual, expected) {
    return chai.expect(actual).to.have.keys(expected);
  },
  'includeKeys': function includeKeys(actual, expected) {
    return chai.expect(actual).to.include.keys(expected);
  },
  'contains': function contains(actual, expected) {
    return chai.expect(actual).to.contain(expected);
  },
  'noterror': function noterror(actual, expected) {
    if (actual != null ? actual.message : void 0) {
      throw actual;
    }
  }
}; // returns a predicate function
// (data) ->

findOperator = function findOperator(expectation) {
  var expectValue, op, opname, predicate;

  for (opname in expectation) {
    expectValue = expectation[opname];

    if (opname === 'path') {
      continue;
    }

    op = operators[opname];

    if (op) {
      predicate = function predicate(data) {
        return op(data, expectValue);
      };

      predicate.toString = function () {
        return "".concat(opname, " ").concat(expectValue);
      };

      return predicate;
    }
  }

  throw new Error("fbp-spec: No operator matching ".concat(Object.keys(expectation), ". Available: ").concat(Object.keys(operators)));
};

extractMatches = function extractMatches(expectation, data) {
  var matches, options;
  options = {
    flatten: true
  };

  if (expectation.path) {
    debug('extracting JSONPath from', expectation.path, data);
    matches = JSONPath.eval(data, expectation.path, options);

    if (!matches.length) {
      throw new Error("expected JSONPath '".concat(expectation.path, "' to match data in ").concat(JSON.stringify(data)));
    }
  } else {
    matches = [data];
  }

  debug('matching against', matches);
  return matches;
};

exports.expect = function (expects, portsdata) {
  var data, e, exp, expectation, expectations, i, len, m, matches, port, predicate, results;

  if (!common.isArray(expects)) {
    // can have one or more set of expectations of messages
    expects = [expects];
  }

  results = [];

  for (i = 0, len = expects.length; i < len; i++) {
    e = expects[i];
    results.push(function () {
      var results1; // each message expectation can match on ports

      results1 = [];

      for (port in e) {
        exp = e[port]; // for each matching port, there can be one or more assertions on the value

        expectations = exp;

        if (!common.isArray(expectations)) {
          expectations = [expectations];
        }

        results1.push(function () {
          var j, len1, results2;
          results2 = [];

          for (j = 0, len1 = expectations.length; j < len1; j++) {
            expectation = expectations[j];
            debug('checking port for expectation', port, expectation);
            data = portsdata[port];

            if (typeof data === 'undefined') {
              throw new Error("No data received on port ".concat(port));
            }

            predicate = findOperator(expectation);
            matches = extractMatches(expectation, data);
            results2.push(function () {
              var k, len2, results3;
              results3 = [];

              for (k = 0, len2 = matches.length; k < len2; k++) {
                m = matches[k];
                debug('checking against predicate', m, predicate.toString());
                results3.push(predicate(m));
              }

              return results3;
            }());
          }

          return results2;
        }());
      }

      return results1;
    }());
  }

  return results;
};

exports.noError = function (maybeErr) {
  return chai.expect(maybeErr).to.not.exist;
};