"use strict";

//# Run Mocha testcases using fbp-spec as a runner/frontend
//# Intended to allow existing Mocha testcases to be seen and executed
//# from a FBP protocol client like Flowhub, without requiring them
//# to be rewritten as fbp-spec tests
//# This is especially useful to allow partial and gradual migration of existing test suites
//# See also ./mocha.coffee, which can be used in combination
// Partially based on example code from https://github.com/mochajs/mocha/wiki/Third-party-UIs
var Mocha,
    buildFbpSpecs,
    debug,
    discoverHost,
    dumpSpecs,
    fbpComponentFromSpec,
    fbpComponentName,
    fbpSourceFromSpec,
    fs,
    handleFbpCommand,
    http,
    knownUnsupportedCommands,
    _loadSuite,
    loadTests,
    main,
    normalizeOptions,
    parse,
    path,
    runTests,
    setup,
    testId,
    testsuite,
    websocket,
    indexOf = [].indexOf;

fs = require('fs');
path = require('path');
http = require('http');
websocket = require('./websocket'); // FIXME: split out transport interface of noflo-runtime-*, use that directly

Mocha = require('mocha');
debug = require('debug')('fbp-spec:mochacompat');
testsuite = require('./testsuite');

loadTests = function loadTests(files) {
  var f, i, len, mocha, options, resolved;
  options = {};
  mocha = new Mocha(options);

  for (i = 0, len = files.length; i < len; i++) {
    f = files[i];
    resolved = require.resolve(f);
    delete require.cache[resolved];
    mocha.addFile(f);
  }

  mocha.loadFiles();
  return mocha;
}; // similar to mocha.run(), but files must be loaded beforehand


runTests = function runTests(mocha, progress, callback) {
  var done, options, registerReporter, reporter, runner, suite;
  suite = mocha.suite;
  options = mocha.options;
  options.files = mocha.files;
  runner = new Mocha.Runner(suite, options.delay);

  registerReporter = function registerReporter(r) {
    runner.on('pass', function (test) {
      return progress(null, test);
    });
    return runner.on('fail', function (test, err) {
      return progress(err, test);
    });
  };

  mocha.reporter(registerReporter, {});
  reporter = new mocha._reporter(runner, options);
  runner.ignoreLeaks = options.ignoreLeaks !== false;
  runner.fullStackTrace = options.fullStackTrace;
  runner.asyncOnly = options.asyncOnly;
  runner.allowUncaught = options.allowUncaught;

  if (options.grep) {
    runner.grep(options.grep, options.invert);
  }

  if (options.globals) {
    runner.globals(options.globals);
  }

  if (options.growl) {
    mocha._growl(runner, reporter);
  }

  if (options.useColors != null) {
    Mocha.reporters.Base.useColors = options.useColors;
  }

  Mocha.reporters.Base.inlineDiffs = options.useInlineDiffs;

  done = function done(failures) {
    if (reporter.done) {
      return reporter.done(failures, callback);
    } else {
      return callback && callback(failures);
    }
  };

  return runner.run(done);
};

testId = function testId(fullname) {
  var crypto, hash;
  crypto = require('crypto');
  hash = crypto.createHash('sha256');
  hash.update(fullname);
  return hash.digest('hex').substr(0, 10);
};

_loadSuite = function loadSuite(fbpSuite, suite) {
  var fbpCase, fullName, i, id, j, len, len1, ref, ref1, results, sub, testcase;
  ref = suite.tests;

  for (i = 0, len = ref.length; i < len; i++) {
    testcase = ref[i]; //console.log 't', testcase

    fullName = fbpSuite.name + testcase.parent.title + testcase.title; // FIXME: make the full name recursive

    id = testId(fullName); // FIXME: salt so it does not collide

    testcase._fbpid = id;
    fbpCase = {
      name: testcase.parent.title,
      assertion: testcase.title,
      _id: id,
      inputs: {
        test: id
      },
      expect: {
        error: {
          noterror: null
        }
      }
    };
    fbpSuite.cases.push(fbpCase);
  }

  ref1 = suite.suites; // load recursively

  results = [];

  for (j = 0, len1 = ref1.length; j < len1; j++) {
    sub = ref1[j];
    results.push(_loadSuite(fbpSuite, sub));
  }

  return results;
};

buildFbpSpecs = function buildFbpSpecs(mocha) {
  var fbpSuite, i, len, ref, specs, suite, top;
  specs = [];
  top = mocha.suite;
  ref = top.suites;

  for (i = 0, len = ref.length; i < len; i++) {
    suite = ref[i]; //console.log 's', suite

    fbpSuite = testsuite.create({
      name: "".concat(suite.title, " (Mocha tests)"),
      fixture: {
        type: 'fbp',
        data: "# @runtime fbp-spec-mocha\n\nINPORT=runTest.IN:TEST\nOUTPORT=runTest.ERROR:ERROR\n\nrunTest(mocha/LoadTestCase) OUT -> IN verifyResult(mocha/CheckResults)"
      }
    });

    _loadSuite(fbpSuite, suite);

    specs.push(fbpSuite);
  }

  return specs;
};

dumpSpecs = function dumpSpecs(suites) {
  var delimiter, i, jsyaml, len, s, str;

  if ((typeof window !== "undefined" && window !== null ? window.jsyaml : void 0) != null) {
    jsyaml = window.jsyaml;
  }

  if (!jsyaml) {
    jsyaml = require('js-yaml');
  }

  str = "";
  delimiter = '---\n';

  for (i = 0, len = suites.length; i < len; i++) {
    s = suites[i];
    str += "".concat(jsyaml.safeDump(s));

    if (suites.length > 1) {
      str += delimiter;
    }
  }

  return str;
};

discoverHost = function discoverHost(preferred_iface) {
  var address, device, filter, ifaces, int_address, os;
  os = require('os'); // node.js only

  ifaces = os.networkInterfaces();
  address = void 0;
  int_address = void 0;

  filter = function filter(connection) {
    if (connection.family !== 'IPv4') {
      return;
    }

    if (connection.internal) {
      int_address = connection.address;
    } else {
      address = connection.address;
    }
  };

  if (typeof preferred_iface === 'string' && indexOf.call(ifaces, preferred_iface) >= 0) {
    ifaces[preferred_iface].forEach(filter);
  } else {
    for (device in ifaces) {
      ifaces[device].forEach(filter);
    }
  }

  return address || int_address;
};

knownUnsupportedCommands = function knownUnsupportedCommands(p, c) {
  return false;
};

fbpComponentName = function fbpComponentName(s) {
  return "fbp-spec-mocha/".concat(s.name);
};

fbpComponentFromSpec = function fbpComponentFromSpec(s) {
  var p; // component:component

  return p = {
    name: fbpComponentName(s),
    subgraph: false,
    inPorts: [],
    outPorts: []
  };
};

fbpSourceFromSpec = function fbpSourceFromSpec(s) {
  var p, serialized; // component:source message, :getsource response

  serialized = dumpSpecs([s]);
  return p = {
    name: fbpComponentName(s),
    code: '',
    language: 'whitespace',
    tests: serialized
  };
};

handleFbpCommand = function handleFbpCommand(state, runtime, mocha, specs, protocol, command, payload, context) {
  var ackMessage, capabilities, checkSendCurrentTest, componentName, found, i, info, j, len, len1, s, testDone, updateStatus;

  updateStatus = function updateStatus(news, event) {
    var runtimeState;

    if (news.started != null) {
      state.started = news.started;
    }

    if (news.running != null) {
      state.running = news.running;
    }

    runtimeState = {
      started: state.started,
      running: state.running,
      graph: payload.graph
    };
    debug('update status', runtimeState);
    return runtime.send('network', event, runtimeState, context);
  }; //sendEvent = (e) ->
  //  runtime.send e.protocol, e.command, e.payload, context


  ackMessage = function ackMessage() {
    // reply with same message as we got in
    return runtime.send(protocol, command, payload, context);
  }; //# Runtime


  if (protocol === 'runtime' && command === 'getruntime') {
    capabilities = ['protocol:graph', // read-only from client
    'protocol:component', // read-only from client
    'protocol:network', 'protocol:runtime', 'component:getsource'];
    info = {
      type: 'fbp-spec-mocha',
      version: '0.5',
      capabilities: capabilities,
      allCapabilities: capabilities,
      graph: 'default/main' // HACK, so Flowhub will ask for our graph

    };
    return runtime.send('runtime', 'runtime', info, context); //sendGraphs mytrace, send, (err) -> # XXX: right place?
    // ignored
  } else if (protocol === 'runtime' && command === 'packet') {
    debug('test message', payload, state.running);

    if (payload.port !== 'test' || payload.event !== 'data') {
      debug('unexpected test message format');
      return;
    }

    state.currentTest = payload.payload; // collect results of completed tests

    testDone = function testDone(err, test) {
      debug('test completed', test._fbpid, err, Object.keys(state.completedTests).length);
      state.completedTests[test._fbpid] = {
        test: test,
        err: err
      };
      return checkSendCurrentTest();
    };

    checkSendCurrentTest = function checkSendCurrentTest() {
      var completed, m;
      completed = state.completedTests[state.currentTest];
      debug('checking for', state.currentTest, completed != null);

      if (completed) {
        m = {
          graph: state.graph,
          event: 'data',
          port: 'error',
          payload: completed.err
        };
        runtime.send('runtime', 'packet', m, context);
        return delete state.completedTests[state.currentTest];
      }
    };

    checkSendCurrentTest(); // we might have completed it already

    if (!state.running) {
      runTests(mocha, testDone, function (f) {
        return updateStatus({
          running: false
        }, 'status');
      });
      return updateStatus({
        running: true
      }, 'status');
    } //# Graph

  } else if (protocol === 'graph' && command === 'addnode') {
    return ackMessage();
  } else if (protocol === 'graph' && command === 'addedge') {
    return ackMessage();
  } else if (protocol === 'graph' && command === 'addinport') {
    return ackMessage();
  } else if (protocol === 'graph' && command === 'addoutport') {
    return ackMessage();
  } else if (protocol === 'graph' && command === 'clear') {
    state.graph = payload.id;
    debug('new graph', state.graph);
    return ackMessage(); //# Network
  } else if (protocol === 'network' && command === 'getstatus') {
    return runtime.send('network', 'status', state, context);
  } else if (protocol === 'network' && command === 'start') {
    debug('FBP network start');
    return updateStatus({
      started: true,
      running: false,
      graph: payload.graph
    }, 'started');
  } else if (protocol === 'network' && command === 'stop') {
    debug('FBP network stop');
    return updateStatus({
      started: false,
      running: false,
      graph: payload.graph
    }, 'stopped'); //# Component
  } else if (protocol === 'component' && command === 'list') {
    // one fake component per Mocha suite
    for (i = 0, len = specs.length; i < len; i++) {
      s = specs[i];
      runtime.send('component', 'component', fbpComponentFromSpec(s), context);
    }

    return runtime.send('component', 'componentsready', specs.length, context);
  } else if (protocol === 'component' && command === 'getsource') {
    // one fake component per Mocha suite
    found = null;

    for (j = 0, len1 = specs.length; j < len1; j++) {
      s = specs[j];
      componentName = fbpComponentName(s);

      if (componentName === payload.name) {
        found = s;
      }
    }

    debug('component getsource', "'".concat(payload.name, "'"), found != null ? found.name : void 0);

    if (found) {
      return runtime.send('component', 'source', fbpSourceFromSpec(found), context);
    }
  } else if (knownUnsupportedCommands(protocol, command)) {} else {
    // ignored
    return debug('Warning: Unknown FBP protocol message', protocol, command);
  }
}; //# Commandline things


normalizeOptions = function normalizeOptions(options) {
  var match;

  if (options.host === 'autodetect') {
    options.host = discoverHost();
  } else if (match = /autodetect\(([a-z0-9]+)\)/.exec(options.host)) {
    options.host = discoverHost(match[1]);
  }

  if (!options.port) {
    options.port = 3333;
  }

  return options;
};

parse = function parse(args) {
  var program;
  program = require('commander'); // TODO: take list of files as input instead, to be more mocha compatible

  program.arguments('<files...>').action(function (args) {
    return program.files = args;
  }).option('--host <hostname>', 'Hostname we serve on, for live-url', String, 'autodetect').option('--port <PORT>', 'Command to launch runtime under test', Number, 3333).parse(process.argv);
  return program;
};

exports.setup = setup = function setup(options, callback) {
  var httpServer, mocha, runtime, specs, state;
  options = normalizeOptions(options);
  mocha = loadTests(options.files);
  specs = buildFbpSpecs(mocha);
  state = {
    started: false,
    running: false,
    currentTest: null,
    graph: null,
    specs: specs,
    completedTests: {} // 'id' -> { err: ?Error, test: MochaTest }

  };
  httpServer = new http.Server();
  runtime = websocket(httpServer, {});

  runtime.receive = function (protocol, command, payload, context) {
    return handleFbpCommand(state, runtime, mocha, specs, protocol, command, payload, context);
  };

  return httpServer.listen(options.port, function (err) {
    return callback(err, state, httpServer);
  });
};

exports.main = main = function main() {
  var options;
  options = parse(process.argv);
  return setup(options, function (err, state) {
    if (err) {
      throw err;
    }

    console.log("fbp-spec-mocha started on ws://".concat(options.host, ":").concat(options.port));
    return console.log("found ".concat(state.specs.length, " test suites"));
  });
};

if (!module.parent) {
  main();
}